#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <memory.h> 
#include <errno.h>
 
struct testdata {
  int port;
  int flag;
  int settwice;
  int expectOk;
};

#define TDSIZE 20
#define BASEPORT 5555

const struct testdata td[TDSIZE] = {
 {0, SO_REUSEADDR, 0, 1}, {0, SO_REUSEADDR, 0, 1}, {0, SO_REUSEPORT, 0, 1},
 {1, SO_REUSEPORT, 0, 1}, {1, SO_REUSEPORT, 0, 1}, {1, SO_REUSEADDR, 0, 1},
 {1, SO_REUSEPORT, 0, 1},
 {2, SO_REUSEADDR, 1, 1}, {2, SO_REUSEADDR, 1, 1}, {2, SO_REUSEPORT, 0, 1},
 {2, SO_REUSEADDR, 0, 1}, {2, SO_REUSEPORT, 0, 1},
 {3, 0, 0, 1}, {3, SO_REUSEADDR, 0, 0}, {3, SO_REUSEPORT, 0, 0},
 {4, SO_REUSEADDR, 0, 1}, {4, 0, 0, 0},
 {5, SO_REUSEPORT, 0, 1}, {5, SO_REUSEPORT, 0, 1}, {5, 0, 0, 0}, 
};

int main(int argc, char *argv[])
{
  int reuse = 1;
  struct sockaddr_in localSock;
  struct sockaddr_in6 localSock6;
  int sd;
  int i;
  int lastport = 0;
  int port;

  memset((char *) &localSock, 0, sizeof(localSock));
  localSock.sin_family = AF_INET;
  localSock.sin_addr.s_addr  = inet_addr("239.1.1.1");

  for (i = 0; i < TDSIZE; ++i) {
    port = BASEPORT + td[i].port;
    localSock.sin_port = htons(port);
    sd = socket(AF_INET, SOCK_DGRAM, 0);
    if (lastport != port) {
      printf("IPv4 Port %i:\n", port);
      lastport = port;
    }
    if (td[i].flag) {
      printf("  Bind using SO_REUSE%s%s...",
            td[i].flag == SO_REUSEADDR ? "ADDR" : "PORT",
            td[i].settwice?" x 2":"....");
      setsockopt(sd, SOL_SOCKET, td[i].flag, (char *)&reuse, sizeof(reuse));
    }
    else {
      printf("  Bind without socketopts.......");
    }
    if (bind(sd, (struct sockaddr*)&localSock, sizeof(localSock))) {
      printf("FAIL (%sexpected): %s\n", td[i].expectOk?"NOT ":"", strerror(errno));
      close(sd);
    }
    else {
      printf("OK   (%sexpected)\n", td[i].expectOk?"":"NOT ");
    }
    if (td[i].settwice && td[i].flag)
     setsockopt(sd, SOL_SOCKET, td[i].flag, (char *)&reuse, sizeof(reuse));
  }

  localSock6.sin6_family = AF_INET6;
  inet_pton(AF_INET6, "ff02::1", &localSock6.sin6_addr.s6_addr);
  for (i = 0; i < TDSIZE; ++i) {
    port = BASEPORT + td[i].port;
    localSock6.sin6_port = htons(port);
    sd = socket(AF_INET6, SOCK_DGRAM, 0);
    if (lastport != port) {
      printf("IPv6 Port %i:\n", port);
      lastport = port;
    }
    if (td[i].flag) {
      printf("  Bind using SO_REUSE%s%s...",
            td[i].flag == SO_REUSEADDR ? "ADDR" : "PORT",
            td[i].settwice?" x 2":"....");
      setsockopt(sd, SOL_SOCKET, td[i].flag, (char *)&reuse, sizeof(reuse));
    }
    else {
      printf("  Bind without socketopts.......");
    }
    if (bind(sd, (struct sockaddr*)&localSock6, sizeof(localSock6))) {
      printf("FAIL (%sexpected): %s\n", td[i].expectOk?"NOT ":"", strerror(errno));
      close(sd);
    }
    else {
      printf("OK   (%sexpected)\n", td[i].expectOk?"":"NOT ");
    }
    if (td[i].settwice && td[i].flag)
     setsockopt(sd, SOL_SOCKET, td[i].flag, (char *)&reuse, sizeof(reuse));
  }
}
